/*
 * Copyright 2008,2010,2011,2012 Red Hat, Inc.
 *
 * This Program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This Program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this Program; if not, write to the
 *
 *   Free Software Foundation, Inc.
 *   59 Temple Place, Suite 330
 *   Boston, MA 02111-1307 USA
 *
 */

#ifndef format_h
#define format_h
struct slapi_entry;
struct slapi_dn;
struct plugin_state;
struct Slapi_ComponentId;
struct Slapi_PluginDesc;

struct format_state;

typedef void (*format_result_callback)(int rc, void *callback_data);
typedef int (*format_referral_entry_callback)(char *referral,
					      void *callback_data);
typedef int (*format_search_entry_callback)(struct slapi_entry *e,
					    void *callback_data);

/* An initialization function that fills in configuration details for the formatter
 * The function has to mark state->initialized as true if configuration is done */
typedef void(*format_state_init_config)(struct format_state *state, Slapi_PBlock *pb);

/* A function to free configuration details for the formatter
 * The function has to mark state->initialized as false if configuration is freed
 * and set set->config values to NULL */
typedef void(*format_state_free_config)(struct format_state *state);

/* A function to return LDAP bases and a filter to search */
typedef void(*format_state_get_bases_filter)(struct format_state *state,
					     const char *group, const char *set,
					     char ***bases, char **filter);
/* A function to free LDAP bases and a filter to search */
typedef void(*format_state_free_bases_filter)(struct format_state *state,
					     char ***bases, char **filter);

/* A callback function similar to slapi_search_internal_callback_pb()
 * in case a search needs to be performed elsewhere */
typedef int(*format_state_search_callback_pb)(struct format_state *state,
					      Slapi_PBlock *pb,
					      void *callback_data,
					      format_result_callback prc,
					      format_search_entry_callback psec,
					      format_referral_entry_callback prec);

struct format_inref_attr {
	char *group, *set, *attribute;
};

struct format_ref_attr_list {
	char *group, *set;
	struct format_ref_attr_list_link {
		char *attribute, *filter_str;
		Slapi_Filter *filter;
		struct slapi_dn **base_sdn_list, **base_sdn_list2;
	} *links;
	int n_links;
};

struct format_set_config {
	const char *disallowed_chars;
	const struct slapi_dn **restrict_subtrees;
	const struct slapi_dn **ignore_subtrees;
	char ***rel_attrs;
	char ***ref_attrs;
	struct format_inref_attr ***inref_attrs;
	struct format_ref_attr_list ***ref_attr_list;
	struct format_ref_attr_list ***inref_attr_list;
};

struct format_state {
	Slapi_ComponentId *plugin_identity;
	Slapi_PluginDesc *plugin_desc;
	void *caller_data;
	bool_t initialized;
	format_state_init_config init_config;
	format_state_free_config free_config;
	format_state_get_bases_filter get_bases_filter;
	format_state_free_bases_filter free_bases_filter;
	format_state_search_callback_pb search_callback_pb;
	struct format_set_config config;
};

struct format_template {
	char *rdn_format;
	char **attribute_format;
	char *disallowed_chars;
};

enum format_flags {
	/* Add operations attributes: only needed if LDAP entry is not stored in LDAP backend */
	FORMAT_OPERATIONAL_ATTRS = 0x0001,
	/* Perform schema check and add extensible object class if the schema check failed */
	FORMAT_EXTENSIBLE_OBJECT = 0x0002,
	/* Reformat resulting entry through slapi_entry2str() / slapi_str2entry()
	 * to remove duplicates, add RDN values, expand object classes.
	 * non-well-formed LDIF is accepted.*/
	FORMAT_REFORMAT_ENTRY = 0x0004,
};

void format_free_attr_list(char **attr_list);
char **format_dup_attr_list(char **attr_list);

void format_free_inref_attrs(struct format_inref_attr **);
struct format_inref_attr **format_dup_inref_attrs(struct format_inref_attr **);

void format_free_ref_attr_list(struct format_ref_attr_list **);
struct format_ref_attr_list **
format_dup_ref_attr_list(struct format_ref_attr_list **);

void format_free_sdn_list(struct slapi_dn **list, struct slapi_dn **list2);
void format_add_sdn_list(struct slapi_dn ***list, struct slapi_dn ***list2,
			 const char *dn);

void format_free_data(char *data);
char *format_get_data(struct plugin_state *state,
		      struct slapi_pblock *pb, struct slapi_entry *e,
		      const char *domain, const char *map,
		      const char *fmt,
		      const char *disallowed_chars,
		      const struct slapi_dn **restrict_subtrees,
		      const struct slapi_dn **ignore_subtrees,
		      char ***rel_attrs,
		      char ***ref_attrs,
		      struct format_inref_attr ***inref_attrs,
		      struct format_ref_attr_list ***ref_attr_list,
		      struct format_ref_attr_list ***inref_attr_list,
		      unsigned int *data_length);
void format_free_data_set(char **data_set, unsigned int *data_lengths);
char **format_get_data_set(struct plugin_state *state,
			   Slapi_PBlock *pb, Slapi_Entry *e,
			   const char *domain, const char *map,
			   const char *fmt,
			   const char *disallowed,
			   const struct slapi_dn **restrict_subtrees,
			   const struct slapi_dn **ignore_subtrees,
			   char ***rel_attrs,
			   char ***ref_attrs,
			   struct format_inref_attr ***inref_attrs,
			   struct format_ref_attr_list ***ref_attr_list,
			   struct format_ref_attr_list ***inref_attr_list,
			   unsigned int **data_lengths);

/* Create `format_state` structure based on the SLAPI plugin-provided details
 * about itself and callbacks for the search and the initialization.
 *
 * Initialization callback should fill in the details of the `format_state.config`
 * structure.
 *
 * Some of the formatting functions require retrieval of configuration of a set
 * in a group to allow look up of entries. `get_bases_filter()` and
 * `free_bases_filter()` provide management of this information.
 *
 * Search callback is used by the formatting functions to request retrieval of the
 * additional LDAP entries if needed. Its behavior should be similar to
 * slapi_search_internall_callback_pb() SLAPI call but it is an abstraction
 * to allow retrieving data from another sources (another 398-ds instances, for
 * example).
 *
 * `caller_data` can be used to pass through additional context to all callbacks
 * as it will be available as `state->caller_data` for them.
 */
struct format_state *
format_create_state(Slapi_ComponentId *plugin_identity,
		    Slapi_PluginDesc *plugin_desc,
		    format_state_init_config init_config,
		    format_state_free_config free_config,
		    format_state_get_bases_filter get_bases_filter,
		    format_state_free_bases_filter free_bases_filter,
		    format_state_search_callback_pb search_callback_pb,
		    void *caller_data);
void format_free_state(struct format_state **state);

/* Return a list of attribute values based on the attribute format `fmt`
 * and original LDAP entry `e`.
 * Length of each value in the list is returned in `data_lengths`, if provided. */
char **format__get_data_set(struct format_state *state,
			    const char *group, const char *set,
			    struct slapi_entry *e,
			    const char *fmt,
			    unsigned int **data_lengths);

/* Return an attribute value based on the attribute format `fmt`
 * and original LDAP entry `e`.
 * Length of the value is returned in `data_length`, if provided. */
char *format__get_data(struct format_state *state,
		       const char *group, const char *set,
		       struct slapi_entry *e,
		       const char *fmt,
		       unsigned int *data_length);

/* Update LDAP entry `target` by applying all formatting directives
 * from the formatting template `format` to an LDAP entry `e`.
 * Returns 0 for success, -1 for errors. */
int format__get_data_entry(struct format_state *state,
			   const char *group, const char *set,
			   struct format_template *format,
			   struct slapi_entry *e,
			   struct slapi_entry *target,
			   enum format_flags flags);

char *format_escape_for_filter(const char *unescaped);
char *format_build_dn(const char *attribute, const char *value,
		      const char *container_sdn);
#endif
